@@ -138,15 +138,19 @@ raptor_ntriples_generate_statement(raptor_parser* parser,
     parser->emitted_default_graph++;
   }
 
-  /* If there is no statement handler - there is nothing else to do */
-  if(!parser->statement_handler)
-    goto cleanup;
-
   statement->subject = subject;
   statement->predicate = predicate;
   statement->object = object;
   statement->graph = graph;
 
+  /* Do not generate a partial triple - but do clean up */
+  if(!subject || !predicate || !object)
+    goto cleanup;
+
+  /* If there is no statement handler - there is nothing else to do */
+  if(!parser->statement_handler)
+    goto cleanup;
+
   /* Generate the statement */
   (*parser->statement_handler)(parser->user_data, statement);
 
@@ -155,288 +159,6 @@ raptor_ntriples_generate_statement(raptor_parser* parser,
 }
 
 
-/* These are for 7-bit ASCII and not locale-specific */
-#define IS_ASCII_ALPHA(c) (((c) > 0x40 && (c) < 0x5B) || ((c) > 0x60 && (c) < 0x7B))
-#define IS_ASCII_UPPER(c) ((c) > 0x40 && (c) < 0x5B)
-#define IS_ASCII_DIGIT(c) ((c) > 0x2F && (c) < 0x3A)
-#define IS_ASCII_PRINT(c) ((c) > 0x1F && (c) < 0x7F)
-#define TO_ASCII_LOWER(c) ((c)+0x20)
-
-typedef enum {
-  RAPTOR_TERM_CLASS_URI,      /* ends on > */
-  RAPTOR_TERM_CLASS_BNODEID,  /* ends on first non [A-Za-z][A-Za-z0-9]* */
-  RAPTOR_TERM_CLASS_STRING,   /* ends on non-escaped " */
-  RAPTOR_TERM_CLASS_LANGUAGE  /* ends on first non [a-z0-9]+ ('-' [a-z0-9]+ )? */
-} raptor_ntriples_term_class;
-
-
-static int 
-raptor_ntriples_term_valid(raptor_parser* rdf_parser,
-                           unsigned char c, int position, 
-                           raptor_ntriples_term_class term_class) 
-{
-  int result = 0;
-
-  switch(term_class) {
-    case RAPTOR_TERM_CLASS_URI:
-      /* ends on > */
-      result = (c != '>');
-      break;
-      
-    case RAPTOR_TERM_CLASS_BNODEID:  
-      /* ends on first non [A-Za-z][A-Za-z0-9]* */
-      result = IS_ASCII_ALPHA(c);
-      if(position)
-        result = (result || IS_ASCII_DIGIT(c));
-      break;
-      
-    case RAPTOR_TERM_CLASS_STRING:
-      /* ends on " */
-      result = (c != '"');
-      break;
-
-    case RAPTOR_TERM_CLASS_LANGUAGE:
-      /* ends on first non [a-zA-Z]+ ('-' [a-zA-Z0-9]+ )? */
-      result = IS_ASCII_ALPHA(c);
-      if(position)
-        result = (result || IS_ASCII_DIGIT(c) || c == '-');
-      break;
-      
-    default:
-      raptor_parser_error(rdf_parser, "Unknown N-Triples term class %d",
-                          term_class);
-  }
-
-  return result;
-}
-
-
-/*
- * raptor_ntriples_term:
- * @parser: NTriples parser
- * @start: pointer to starting character of string (in)
- * @dest: destination of string (in)
- * @lenp: pointer to length of string (in/out)
- * @dest_lenp: pointer to length of destination string (out)
- * @end_char: string ending character
- * @class: string class
- *
- * Parse an N-Triples term with escapes.
- *
- * Relies that @dest is long enough; it need only be as large as the
- * input string @start since when UTF-8 encoding, the escapes are
- * removed and the result is always less than or equal to length of
- * input.
- *
- * N-Triples strings/URIs are written in ASCII at present; characters
- * outside the printable ASCII range are discarded with a warning.
- * See the grammar for full details of the allowed ranges.
- *
- * UTF-8 and the \u and \U esapes are both allowed.
- *
- * Return value: Non 0 on failure
- **/
-static int
-raptor_ntriples_term(raptor_parser* rdf_parser, 
-                     const unsigned char **start, unsigned char *dest, 
-                     size_t *lenp, size_t *dest_lenp,
-                     char end_char,
-                     raptor_ntriples_term_class term_class)
-{
-  const unsigned char *p = *start;
-  unsigned char c = '\0';
-  size_t ulen = 0;
-  unsigned long unichar = 0;
-  unsigned int position = 0;
-  int end_char_seen = 0;
-
-  /* find end of string, fixing backslashed characters on the way */
-  while(*lenp > 0) {
-    int unichar_width;
-
-    c = *p;
-
-    p++;
-    (*lenp)--;
-    rdf_parser->locator.column++;
-    rdf_parser->locator.byte++;
-
-    if(c > 0x7f) {
-      /* just copy the UTF-8 bytes through */
-      int unichar_len;
-      unichar_len = raptor_unicode_utf8_string_get_char(p - 1, 1 + *lenp, NULL);
-      if(unichar_len < 0 || RAPTOR_GOOD_CAST(size_t, unichar_len) > *lenp) {
-        raptor_parser_error(rdf_parser, "UTF-8 encoding error at character %d (0x%02X) found.", c, c);
-        /* UTF-8 encoding had an error or ended in the middle of a string */
-        return 1;
-      }
-      memmove(dest, p-1, unichar_len);
-      dest += unichar_len;
-
-      unichar_len--; /* p, *lenp were moved on by 1 earlier */
-      
-      p += unichar_len;
-      (*lenp) -= unichar_len;
-      rdf_parser->locator.column += unichar_len;
-      rdf_parser->locator.byte += unichar_len;
-      continue;
-    }
-    
-    if(c != '\\') {
-      /* finish at non-backslashed end_char */
-      if(end_char && c == end_char) {
-        end_char_seen = 1;
-        break;
-      }
-
-      if(!raptor_ntriples_term_valid(rdf_parser, c, position, term_class)) {
-        if(end_char) {
-          /* end char was expected, so finding an invalid thing is an error */
-          raptor_parser_error(rdf_parser, "Missing terminating '%c' (found '%c')", end_char, c);
-          return 0;
-        } else {
-          /* it's the end - so rewind 1 to save next char */
-          p--;
-          (*lenp)++;
-          rdf_parser->locator.column--;
-          rdf_parser->locator.byte--;
-          break;
-        }
-      }
-      
-      /* otherwise store and move on */
-      *dest++ = c;
-      position++;
-      continue;
-    }
-
-    if(!*lenp) {
-      raptor_parser_error(rdf_parser, "\\ at end of line");
-      return 0;
-    }
-
-    c = *p;
-
-    p++;
-    (*lenp)--;
-    rdf_parser->locator.column++;
-    rdf_parser->locator.byte++;
-
-    switch(c) {
-      case '"':
-      case '\\':
-        *dest++ = c;
-        break;
-      case 'n':
-        *dest++ = '\n';
-        break;
-      case 'r':
-        *dest++ = '\r';
-        break;
-      case 't':
-        *dest++ = '\t';
-        break;
-      case '<':
-      case '>':
-      case '{':
-      case '}':
-      case '|':
-      case '^':
-      case '`':
-        /* Turtle 2013 allows these in URIs (as well as \" and \\) */
-        *dest++ = c;
-        break;
-
-      case 'u':
-      case 'U':
-        ulen = (c == 'u') ? 4 : 8;
-        
-        if(*lenp < ulen) {
-          raptor_parser_error(rdf_parser, "%c over end of line", c);
-          return 0;
-        }
-
-        if(1) {
-          unsigned int ii;
-          int n = 0;
-
-          for(ii = 0; ii < ulen; ii++) {
-            char cc = p[ii];
-            if(!isxdigit(RAPTOR_GOOD_CAST(char, cc))) {
-              raptor_parser_error(rdf_parser,
-                            "N-Triples string error - illegal hex digit %c in Unicode escape '%c%s...'",
-                            cc, c, p);
-              n = 1;
-              break;
-            }
-          }
-
-          if(n)
-            break;
-          
-          n = sscanf((const char*)p, ((ulen == 4) ? "%04lx" : "%08lx"), &unichar);
-          if(n != 1) {
-            raptor_parser_error(rdf_parser, "Illegal Uncode escape '%c%s...'", c, p);
-            break;
-          }
-        }
-
-        p += ulen;
-        (*lenp) -= ulen;
-        rdf_parser->locator.column += RAPTOR_GOOD_CAST(int, ulen);
-        rdf_parser->locator.byte += RAPTOR_GOOD_CAST(int, ulen);
-        
-        if(unichar > raptor_unicode_max_codepoint) {
-          raptor_parser_error(rdf_parser,
-                              "Illegal Unicode character with code point #x%lX (max #x%lX).",
-                              unichar, raptor_unicode_max_codepoint);
-          break;
-        }
-
-        unichar_width = raptor_unicode_utf8_string_put_char(unichar, dest, 4);
-        if(unichar_width < 0) {
-          raptor_parser_error(rdf_parser,
-                              "Illegal Unicode character with code point #x%lX.", 
-                              unichar);
-          break;
-        }
-
-        /* The destination length is set here to 4 since we know that in 
-         * all cases, the UTF-8 encoded output sequence is always shorter
-         * than the input sequence, and the buffer is edited in place.
-         *   \uXXXX: 6 bytes input - UTF-8 max 3 bytes output
-         *   \uXXXXXXXX: 10 bytes input - UTF-8 max 4 bytes output
-         */
-        dest += (int)unichar_width;
-        break;
-
-      default:
-        raptor_parser_error(rdf_parser,
-                            "Illegal string escape \\%c in \"%s\"", c, 
-                            (char*)start);
-        return 0;
-    }
-
-    position++;
-  } /* end while */
-
-  
-  if(end_char && !end_char_seen) {
-    raptor_parser_error(rdf_parser, "Missing terminating '%c' before end of line.", end_char);
-    return 1;
-  }
-
-  /* terminate dest, can be shorter than source */
-  *dest = '\0';
-
-  if(dest_lenp)
-    *dest_lenp = p - *start;
-
-  *start = p;
-
-  return 0;
-}
-
 
 #define MAX_NTRIPLES_TERMS 4
 
@@ -448,9 +170,7 @@ raptor_ntriples_parse_line(raptor_parser* rdf_parser,
   raptor_ntriples_parser_context *ntriples_parser = (raptor_ntriples_parser_context*)rdf_parser->context;
   int i;
   unsigned char *p;
-  unsigned char *dest;
-  raptor_term* real_terms[MAX_NTRIPLES_TERMS] = {NULL, NULL, NULL, NULL};
-  size_t term_length = 0;
+  raptor_term* terms[MAX_NTRIPLES_TERMS+1] = {NULL, NULL, NULL, NULL, NULL};
   int rc = 0;
   
   /* ASSERTION:
@@ -493,259 +213,68 @@ raptor_ntriples_parse_line(raptor_parser* rdf_parser,
   
   /* Must be triple/quad */
 
-  for(i = 0; i < 4; i++) {
+  for(i = 0; i < MAX_NTRIPLES_TERMS + 1; i++) {
     if(!len) {
-      /* context is optional in nquads */
-      if (i == 3)
-	break;
+      if(ntriples_parser->is_nquads) {
+        /* context is optional in nquads */
+        if(i == 3 || i ==4)
+          break;
+      } else {
+        if(i == 3)
+          break;
+      }
       raptor_parser_error(rdf_parser, "Unexpected end of line");
       goto cleanup;
     }
     
 
-    if(i == 2) {
+    if(i == 3) {
+      /* graph term (3): blank node or <URI> */
+      if(*p != '<' && *p != '_') {
+        raptor_parser_error(rdf_parser, "Saw '%c', expected Graph term <URIref>, _:bnodeID", *p);
+        goto cleanup;
+      }
+    } else if(i == 2) {
       /* object term (2): expect either <URI> or _:name or literal */
-      if(*p != '<' && *p != '_' && *p != '"' && *p != 'x') {
-        raptor_parser_error(rdf_parser, "Saw '%c', expected <URIref>, _:bnodeID or \"literal\"", *p);
+      if(*p != '<' && *p != '_' && *p != '"') {
+        raptor_parser_error(rdf_parser, "Saw '%c', expected object term <URIref>, _:bnodeID or \"literal\"", *p);
         goto cleanup;
       }
     } else if(i == 1) {
       /* predicate term (1): expect URI only */
       if(*p != '<') {
-        raptor_parser_error(rdf_parser, "Saw '%c', expected <URIref>", *p);
+        raptor_parser_error(rdf_parser, "Saw '%c', expected predict term <URIref>", *p);
         goto cleanup;
       }
     } else {
       /* subject (0) or graph (3) terms: expect <URI> or _:name */
       if(*p != '<' && *p != '_') {
-        raptor_parser_error(rdf_parser, "Saw '%c', expected <URIref> or _:bnodeID", *p);
+        raptor_parser_error(rdf_parser, "Saw '%c', expected subject term <URIref> or _:bnodeID", *p);
         goto cleanup;
       }
     }
 
 
-    switch(*p) {
-      case '<':
-        dest = p;
-
-        p++;
-        len--;
-        rdf_parser->locator.column++;
-        rdf_parser->locator.byte++;
-
-        if(raptor_ntriples_term(rdf_parser,
-                                (const unsigned char**)&p, 
-                                dest, &len, &term_length, 
-                                '>', RAPTOR_TERM_CLASS_URI)) {
-          rc = 1;
-          goto cleanup;
-        }
-
-        if(!raptor_turtle_check_uri_string(dest)) {
-          raptor_parser_error(rdf_parser, "URI '%s' contains bad character(s)", dest);
-          rc = 1;
-          goto cleanup;
-        }
-
-        if(1) {
-          raptor_uri *uri;
-
-          /* Check for bad ordinal predicate */
-          if(!strncmp((const char*)dest,
-                      "http://www.w3.org/1999/02/22-rdf-syntax-ns#_", 44)) {
-            int ordinal = raptor_check_ordinal(dest + 44);
-            if(ordinal <= 0)
-              raptor_parser_error(rdf_parser, "Illegal ordinal value %d in property '%s'.", ordinal, dest);
-          }
-          if(raptor_uri_uri_string_is_absolute(dest) <= 0) {
-            raptor_parser_error(rdf_parser, "URI '%s' is not absolute.", dest);
-            goto cleanup;
-          }
-          
-          uri = raptor_new_uri(rdf_parser->world, dest);
-          if(!uri) {
-            raptor_parser_error(rdf_parser, "Could not create URI for '%s'", (const char *)dest);
-            goto cleanup;
-          }
-          real_terms[i] = raptor_new_term_from_uri(rdf_parser->world, uri);
-          raptor_free_uri(uri);
-        }
-        break;
-
-      case '"':
-        dest = p;
-
-        p++;
-        len--;
-        rdf_parser->locator.column++;
-        rdf_parser->locator.byte++;
-
-        if(raptor_ntriples_term(rdf_parser,
-                                (const unsigned char**)&p,
-                                dest, &len, &term_length,
-                                '"', RAPTOR_TERM_CLASS_STRING)) {
-          rc = 1;
-          goto cleanup;
-        }
-
-        if(1) {
-          unsigned char *object_literal_language = NULL;
-          unsigned char *object_literal_datatype = NULL;
-          raptor_uri* datatype_uri = NULL;
+    rc = raptor_ntriples_parse_term(rdf_parser->world, &rdf_parser->locator,
+                                    p, &len, &terms[i], 0);
     
-          if(len && *p == '@') {
-            unsigned char *q;
-            size_t lang_len;
-            
-            object_literal_language = p;
-
-            /* Skip - */
-            p++;
-            len--;
-            rdf_parser->locator.column++;
-            rdf_parser->locator.byte++;
-
-            if(!len) {
-              raptor_parser_error(rdf_parser, "Missing language after \"string\"-");
-              goto cleanup;
-            }
-
-
-            if(raptor_ntriples_term(rdf_parser,
-                                    (const unsigned char**)&p,
-                                    object_literal_language, &len, &lang_len,
-                                    '\0', RAPTOR_TERM_CLASS_LANGUAGE)) {
-              rc = 1;
-              goto cleanup;
-            }
-
-            if(!lang_len) {
-              raptor_parser_error(rdf_parser, "Invalid language tag at @%s", p);
-              rc = 1;
-              goto cleanup;
-            }
-            
-            /* Normalize language to lowercase
-             * http://www.w3.org/TR/rdf-concepts/#dfn-language-identifier
-             */
-            for(q = object_literal_language; *q; q++) {
-              if(IS_ASCII_UPPER(*q))
-                *q = TO_ASCII_LOWER(*q);
-            }
-
-          }
-
-          if(len >1 && *p == '^' && p[1] == '^') {
-
-            object_literal_datatype = p;
-
-            /* Skip ^^ */
-            p += 2;
-            len -= 2;
-            rdf_parser->locator.column += 2;
-            rdf_parser->locator.byte += 2;
-
-            if(!len || (len && *p != '<')) {
-              raptor_parser_error(rdf_parser, "Missing datatype URI-ref in\"string\"^^<URI-ref> after ^^");
-              goto cleanup;
-            }
-
-            p++;
-            len--;
-            rdf_parser->locator.column++;
-            rdf_parser->locator.byte++;
-
-            if(raptor_ntriples_term(rdf_parser,
-                                    (const unsigned char**)&p,
-                                    object_literal_datatype, &len, NULL,
-                                    '>', RAPTOR_TERM_CLASS_URI)) {
-              rc = 1;
-              goto cleanup;
-            }
-
-            if(raptor_uri_uri_string_is_absolute(object_literal_datatype) <= 0) {
-              raptor_parser_error(rdf_parser, "Datatype URI '%s' is not absolute.", object_literal_datatype);
-              rc = 1;
-              goto cleanup;
-            }
-
-          }
-
-          if(object_literal_datatype && object_literal_language) {
-            raptor_parser_warning(rdf_parser, "Typed literal used with a language - ignoring the language");
-            object_literal_language = NULL;
-          }
-
-          if(object_literal_datatype) {
-            datatype_uri = raptor_new_uri(rdf_parser->world,
-                                          object_literal_datatype);
-            if(!datatype_uri) {
-              raptor_parser_error(rdf_parser, "Could not create literal datatype uri '%s'", object_literal_datatype);
-              goto cleanup;
-            }
-            object_literal_language = NULL;
-          }
-          
-          real_terms[i] = raptor_new_term_from_literal(rdf_parser->world,
-                                                       dest,
-                                                       datatype_uri,
-                                                       object_literal_language);
-        }
-
-        break;
-
-
-      case '_':
-        /* store where _ was */
-        dest = p;
-
-        p++;
-        len--;
-        rdf_parser->locator.column++;
-        rdf_parser->locator.byte++;
-
-        if(!len || (len > 0 && *p != ':')) {
-          raptor_parser_error(rdf_parser, "Illegal bNodeID - _ not followed by :");
-          goto cleanup;
-        }
-
-        /* Found ':' - move on */
-
-        p++;
-        len--;
-        rdf_parser->locator.column++;
-        rdf_parser->locator.byte++;
-
-        if(raptor_ntriples_term(rdf_parser,
-                                (const unsigned char**)&p,
-                                dest, &len, &term_length,
-                                '\0', RAPTOR_TERM_CLASS_BNODEID)) {
-          rc = 1;
-          goto cleanup;
-        }
-
-        if(!term_length) {
-          raptor_parser_error(rdf_parser, "Bad or missing bNodeID after _:");
-          goto cleanup;
-        }
-
-        real_terms[i] = raptor_new_term_from_blank(rdf_parser->world, dest);
-
-        break;
-
-      default:
-        raptor_parser_fatal_error(rdf_parser, "Unknown term type");
-        rc = 1;
-        goto cleanup;
-    }
-
-
-    /* Whitespace must separate the terms */
-    if(i < 2 && !isspace((int)*p)) {
-      raptor_parser_error(rdf_parser, "Missing whitespace after term");
+    if(!rc) {
       rc = 1;
       goto cleanup;
     }
+    p += rc;
+    rc = 0;
+
+    if(terms[i] && terms[i]->type == RAPTOR_TERM_TYPE_URI) {
+      unsigned const char* uri_string;
+
+      /* Check for absolute URI */
+      uri_string = raptor_uri_as_string(terms[i]->value.uri);
+      if(!raptor_uri_uri_string_is_absolute(uri_string)) {
+        raptor_parser_error(rdf_parser, "URI %s is not absolute", uri_string);
+        goto cleanup;
+      }
+    }
 
     /* Skip whitespace after terms */
     while(len > 0 && isspace((int)*p)) {
@@ -756,12 +285,13 @@ raptor_ntriples_parse_line(raptor_parser* rdf_parser,
     }
 
 #if defined(RAPTOR_DEBUG) && RAPTOR_DEBUG > 1
-    if(1) {
-      unsigned char* c = raptor_term_to_string(real_terms[i]);
+    if(terms[i]) {
+      unsigned char* c = raptor_term_to_string(terms[i]);
       fprintf(stderr, "item %d: term '%s' type %d\n",
-              i, c, real_terms[i]->type);
-      raptor_free_string(c);
-    }
+              i, c, terms[i]->type);
+      raptor_free_memory(c);
+    } else
+      fprintf(stderr, "item %d: NULL term\n", i);
 #endif
 
     /* Look for terminating '.' after 3rd (ntriples) or 3rd/4th (nquads) term */
@@ -796,29 +326,40 @@ raptor_ntriples_parse_line(raptor_parser* rdf_parser,
   }
 
 
-  /* Check N-Triples has only 3 terms */
-  if(!ntriples_parser->is_nquads) {
-    if(real_terms[3]) {
-      raptor_free_term(real_terms[3]);
-      real_terms[3] = NULL;
+  if(ntriples_parser->is_nquads) {
+    /* Check N-Quads has 3 or 4 terms */
+    if(terms[4]) {
+      raptor_free_term(terms[4]);
+      terms[4] = NULL;
+      raptor_parser_error(rdf_parser, "N-Quads only allows 3 or 4 terms");
+      goto cleanup;
+    }
+  } else {
+    /* Check N-Triples has only 3 terms */
+    if(terms[3] || terms[4]) {
+      if(terms[4]) {
+        raptor_free_term(terms[4]);
+        terms[4] = NULL;
+      }
+      if(terms[3]) {
+        raptor_free_term(terms[3]);
+        terms[3] = NULL;
+      }
       raptor_parser_error(rdf_parser, "N-Triples only allows 3 terms");
       goto cleanup;
     }
   }
 
-  if(real_terms[3] && real_terms[3]->type == RAPTOR_TERM_TYPE_LITERAL) {
+  if(terms[3] && terms[3]->type == RAPTOR_TERM_TYPE_LITERAL) {
     if(!ntriples_parser->literal_graph_warning++)
       raptor_parser_warning(rdf_parser, "Ignoring N-Quad literal contexts");
 
-    raptor_free_term(real_terms[3]);
-    real_terms[3] = NULL;
+    raptor_free_term(terms[3]);
+    terms[3] = NULL;
   }
 
   raptor_ntriples_generate_statement(rdf_parser, 
-                                     real_terms[0],
-                                     real_terms[1],
-                                     real_terms[2],
-                                     real_terms[3]);
+                                     terms[0], terms[1], terms[2], terms[3]);
 
   rdf_parser->locator.byte += RAPTOR_BAD_CAST(int, len);
 
@@ -838,49 +379,54 @@ raptor_ntriples_parse_chunk(raptor_parser* rdf_parser,
   unsigned char *start;
   raptor_ntriples_parser_context *ntriples_parser = (raptor_ntriples_parser_context*)rdf_parser->context;
   int max_terms = ntriples_parser->is_nquads ? 4 : 3;
-  
+  unsigned char* end_ptr;
+
 #if defined(RAPTOR_DEBUG) && RAPTOR_DEBUG > 1
   RAPTOR_DEBUG2("adding %d bytes to buffer\n", (unsigned int)len);
 #endif
 
-  /* No data?  It's the end */
-  if(!len)
-    return 0;
+  if(len) {
+    buffer = RAPTOR_MALLOC(unsigned char*, ntriples_parser->line_length + len + 1);
+    if(!buffer) {
+      raptor_parser_fatal_error(rdf_parser, "Out of memory");
+      return 1;
+    }
 
-  buffer = RAPTOR_MALLOC(unsigned char*, ntriples_parser->line_length + len + 1);
-  if(!buffer) {
-    raptor_parser_fatal_error(rdf_parser, "Out of memory");
-    return 1;
-  }
+    if(ntriples_parser->line_length) {
+      memcpy(buffer, ntriples_parser->line, ntriples_parser->line_length);
+      RAPTOR_FREE(char*, ntriples_parser->line);
+    }
 
-  if(ntriples_parser->line_length) {
-    memcpy(buffer, ntriples_parser->line, ntriples_parser->line_length);
-    RAPTOR_FREE(char*, ntriples_parser->line);
-  }
+    ntriples_parser->line = buffer;
 
-  ntriples_parser->line = buffer;
+    /* move pointer to end of cdata buffer */
+    ptr = buffer + ntriples_parser->line_length;
 
-  /* move pointer to end of cdata buffer */
-  ptr = buffer+ntriples_parser->line_length;
+    /* adjust stored length */
+    ntriples_parser->line_length += len;
 
-  /* adjust stored length */
-  ntriples_parser->line_length += len;
+    /* now write new stuff at end of cdata buffer */
+    memcpy(ptr, s, len);
+    ptr += len;
+    *ptr = '\0';
+  } else
+    buffer = ntriples_parser->line;
 
-  /* now write new stuff at end of cdata buffer */
-  memcpy(ptr, s, len);
-  ptr += len;
-  *ptr = '\0';
 
 #if defined(RAPTOR_DEBUG) && RAPTOR_DEBUG > 1
-  RAPTOR_DEBUG2("buffer now %d bytes\n", ntriples_parser->line_length);
+  RAPTOR_DEBUG2("buffer now %ld bytes\n", ntriples_parser->line_length);
 #endif
 
-  ptr = buffer+ntriples_parser->offset;
-  while(*(start = ptr)) {
+  if(!ntriples_parser->line_length)
+    return 0;
+
+  ptr = buffer + ntriples_parser->offset;
+  end_ptr = buffer + ntriples_parser->line_length;
+  while((start = ptr) < end_ptr) {
     unsigned char *line_start = ptr;
-    
+
 #if defined(RAPTOR_DEBUG) && RAPTOR_DEBUG > 1
-  RAPTOR_DEBUG3("line buffer now '%s' (offset %d)\n", ptr, ptr-(buffer+ntriples_parser->offset));
+  RAPTOR_DEBUG3("line buffer now '%s' (offset %ld)\n", ptr, ptr-(buffer+ntriples_parser->offset));
 #endif
 
     /* skip \n when just seen \r - i.e. \r\n or CR LF */
@@ -894,21 +440,52 @@ raptor_ntriples_parse_chunk(raptor_parser* rdf_parser,
       start = line_start = ptr;
     }
 
-    while(*ptr && *ptr != '\n' && *ptr != '\r')
-      ptr++;
+    if(1) {
+      int quote = '\0';
+      int bq = 0;
+      while(ptr < end_ptr) {
+        if(!bq) {
+          if(*ptr == '\\') {
+            bq = 1;
+            ptr++;
+            continue;
+          }
 
-    if(!*ptr)
-      break;
+          if(!quote) {
+            if(*ptr == '\'' || *ptr == '"')
+              quote = *ptr;
+            if(*ptr == '\n' || *ptr == '\r')
+              break;
+          } else {
+            if(*ptr == quote)
+              quote = 0;
+          }
+        }
+        ptr++;
+        bq = 0;
+      }
+    }
 
+    if(ptr == end_ptr) {
+      if(!is_end)
+        /* middle of line */
+        break;
+    } else {
 #if defined(RAPTOR_DEBUG) && RAPTOR_DEBUG > 1
-    RAPTOR_DEBUG3("found newline \\x%02x at offset %d\n", *ptr,
-                  ptr-line_start);
+      RAPTOR_DEBUG3("found newline \\x%02x at offset %ld\n", *ptr,
+                    ptr-line_start);
 #endif
-    ntriples_parser->last_char = *ptr;
-
-    len = ptr-line_start;
+      ntriples_parser->last_char = *ptr;
+    }
+    
+    len = ptr - line_start;
     rdf_parser->locator.column = 0;
 
+#if defined(RAPTOR_DEBUG) && RAPTOR_DEBUG > 1
+    RAPTOR_DEBUG2("line (%ld) : >>>", len);
+    fwrite(line_start, sizeof(char), len, stderr);
+    fputs("<<<\n", stderr);
+#endif
     *ptr = '\0';
     if(raptor_ntriples_parse_line(rdf_parser, line_start, len, max_terms))
       return 1;
@@ -916,12 +493,14 @@ raptor_ntriples_parse_chunk(raptor_parser* rdf_parser,
     rdf_parser->locator.line++;
 
     /* go past newline */
-    ptr++;
-    rdf_parser->locator.byte++;
+    if(ptr < end_ptr) {
+      ptr++;
+      rdf_parser->locator.byte++;
+    }
 
 #if defined(RAPTOR_DEBUG) && RAPTOR_DEBUG > 1
     /* Do not peek if too far */
-    if(ptr-buffer < ntriples_parser->line_length)
+    if(RAPTOR_BAD_CAST(size_t, ptr - buffer) < ntriples_parser->line_length)
       RAPTOR_DEBUG2("next char is \\x%02x\n", *ptr);
     else
       RAPTOR_DEBUG1("next char unknown - end of buffer\n");
@@ -932,11 +511,11 @@ raptor_ntriples_parse_chunk(raptor_parser* rdf_parser,
 
   len = ntriples_parser->line_length - ntriples_parser->offset;
     
-  if(len) {
+  if(len && ntriples_parser->line_length != len) {
     /* collapse buffer */
 
 #if defined(RAPTOR_DEBUG) && RAPTOR_DEBUG > 1
-    RAPTOR_DEBUG3("collapsing buffer from %d to %d bytes\n", ntriples_parser->line_length, (unsigned int)len);
+    RAPTOR_DEBUG3("collapsing buffer from %ld to %ld bytes\n", ntriples_parser->line_length, len);
 #endif
     buffer = RAPTOR_MALLOC(unsigned char*, len + 1);
     if(!buffer) {
@@ -956,14 +535,14 @@ raptor_ntriples_parse_chunk(raptor_parser* rdf_parser,
     ntriples_parser->offset = 0;
 
 #if defined(RAPTOR_DEBUG) && RAPTOR_DEBUG > 1
-    RAPTOR_DEBUG3("buffer now '%s' (%d bytes)\n", ntriples_parser->line, ntriples_parser->line_length);
+    RAPTOR_DEBUG3("buffer now '%s' (%ld bytes)\n", ntriples_parser->line, ntriples_parser->line_length);
 #endif    
   }
-  
+
   /* exit now, no more input */
   if(is_end) {
     if(ntriples_parser->offset != ntriples_parser->line_length) {
-       raptor_parser_error(rdf_parser, "Junk at end of input.\"");
+       raptor_parser_error(rdf_parser, "Junk at end of input.");
        return 1;
     }
     
