@@ -27,11 +27,6 @@
 #include <raptor_config.h>
 #endif
 
-#ifdef WIN32
-#include <win32_raptor_config.h>
-#endif
-
-
 #include <stdio.h>
 #include <string.h>
 #include <ctype.h>
@@ -47,12 +42,15 @@
 #include "raptor2.h"
 #include "raptor_internal.h"
 
+#ifdef MAINTAINER_MODE
+#include <git-version.h>
+#endif
 
 /* statics */
 
-const char * const raptor_short_copyright_string = "Copyright 2000-2011 David Beckett. Copyright 2000-2005 University of Bristol";
+const char * const raptor_short_copyright_string = "Copyright 2000-2013 David Beckett. Copyright 2000-2005 University of Bristol";
 
-const char * const raptor_copyright_string = "Copyright (C) 2000-2011 David Beckett - http://www.dajobe.org/\nCopyright (C) 2000-2005 University of Bristol - http://www.bristol.ac.uk/";
+const char * const raptor_copyright_string = "Copyright (C) 2000-2013 David Beckett - http://www.dajobe.org/\nCopyright (C) 2000-2005 University of Bristol - http://www.bristol.ac.uk/";
 
 const char * const raptor_license_string = "LGPL 2.1 or newer, GPL 2 or newer, Apache 2.0 or newer.\nSee http://librdf.org/raptor/LICENSE.html for full terms.";
 
@@ -65,7 +63,7 @@ const char * const raptor_home_url_string = "http://librdf.org/raptor/";
  *
  * See also #raptor_version_decimal.
  */
-const char * const raptor_version_string = VERSION
+const char * const raptor_version_string = RAPTOR_VERSION_STRING
 #ifdef GIT_VERSION
 " GIT " GIT_VERSION
 #endif
