@@ -76,7 +76,8 @@ void raptor_sign_free(void *ptr);
 
 #endif
 
-#ifndef __func__
+#ifdef HAVE___FUNC__
+#else
 #define __func__ "???"
 #endif
 
@@ -691,6 +692,9 @@ void raptor_stats_print(raptor_parser *rdf_parser, FILE *stream);
 RAPTOR_INTERNAL_API const char* raptor_basename(const char *name);
 int raptor_term_print_as_ntriples(const raptor_term *term, FILE* stream);
 
+/* raptor_ntriples.c */
+size_t raptor_ntriples_parse_term(raptor_world* world, raptor_locator* locator, unsigned char *string, size_t *len_p, raptor_term** term_p, int allow_turtle);
+
 /* raptor_parse.c */
 raptor_parser_factory* raptor_world_get_parser_factory(raptor_world* world, const char *name);  
 void raptor_delete_parser_factories(void);
@@ -1432,6 +1436,12 @@ struct raptor_world_s {
   int default_generate_bnodeid_handler_base;
   char *default_generate_bnodeid_handler_prefix;
   unsigned int default_generate_bnodeid_handler_prefix_length;
+
+  raptor_uri* xsd_namespace_uri;
+  raptor_uri* xsd_boolean_uri;
+  raptor_uri* xsd_decimal_uri;
+  raptor_uri* xsd_double_uri;
+  raptor_uri* xsd_integer_uri;
 };
 
 /* raptor_www.c */
